@@ -17,19 +17,22 @@ __all__ = [
     'A', 'A6', 'AAAA', 'AFSDB', 'CNAME', 'DNAME', 'HINFO',
     'MAILA', 'MAILB', 'MB', 'MD', 'MF', 'MG', 'MINFO', 'MR', 'MX',
     'NAPTR', 'NS', 'NULL', 'OPT', 'PTR', 'RP', 'SOA', 'SPF', 'SRV', 'TXT',
-    'WKS',
+    'SSHFP', 'TSIG', 'WKS',
 
     'ANY', 'CH', 'CS', 'HS', 'IN',
 
     'ALL_RECORDS', 'AXFR', 'IXFR',
 
     'EFORMAT', 'ENAME', 'ENOTIMP', 'EREFUSED', 'ESERVER', 'EBADVERSION',
+    'EBADSIG', 'EBADKEY', 'EBADTIME',
 
     'Record_A', 'Record_A6', 'Record_AAAA', 'Record_AFSDB', 'Record_CNAME',
     'Record_DNAME', 'Record_HINFO', 'Record_MB', 'Record_MD', 'Record_MF',
     'Record_MG', 'Record_MINFO', 'Record_MR', 'Record_MX', 'Record_NAPTR',
     'Record_NS', 'Record_NULL', 'Record_PTR', 'Record_RP', 'Record_SOA',
-    'Record_SPF', 'Record_SRV', 'Record_TXT', 'Record_WKS', 'UnknownRecord',
+    'Record_SPF', 'Record_SRV', 'Record_SSHFP', 'Record_TSIG', 'Record_TXT',
+    'Record_WKS',
+    'UnknownRecord',
 
     'QUERY_CLASSES', 'QUERY_TYPES', 'REV_CLASSES', 'REV_TYPES', 'EXT_QUERIES',
 
@@ -104,7 +107,8 @@ else:
 
 def randomSource():
     """
-    Wrapper around L{randbytes.secureRandom} to return 2 random chars.
+    Wrapper around L{twisted.python.randbytes.RandomFactory.secureRandom} to return
+    2 random chars.
     """
     return struct.unpack('H', randbytes.secureRandom(2, fallback=True))[0]
 
@@ -119,8 +123,14 @@ NAPTR = 35
 A6 = 38
 DNAME = 39
 OPT = 41
+SSHFP = 44
 SPF = 99
 
+# These record types do not exist in zones, but are transferred in
+# messages the same way normal RRs are.
+TKEY = 249
+TSIG = 250
+
 QUERY_TYPES = {
     A: 'A',
     NS: 'NS',
@@ -149,7 +159,11 @@ QUERY_TYPES = {
     A6: 'A6',
     DNAME: 'DNAME',
     OPT: 'OPT',
-    SPF: 'SPF'
+    SSHFP: 'SSHFP',
+    SPF: 'SPF',
+
+    TKEY: 'TKEY',
+    TSIG: 'TSIG',
 }
 
 IXFR, AXFR, MAILB, MAILA, ALL_RECORDS = range(251, 256)
@@ -192,10 +206,14 @@ OP_UPDATE = 5 # RFC 2136
 OK, EFORMAT, ESERVER, ENAME, ENOTIMP, EREFUSED = range(6)
 # https://tools.ietf.org/html/rfc6891#section-9
 EBADVERSION = 16
+# RFC 2845
+EBADSIG, EBADKEY, EBADTIME = range(16, 19)
+
+
 
 class IRecord(Interface):
     """
-    An single entry in a zone of authority.
+    A single entry in a zone of authority.
     """
 
     TYPE = Attribute("An indicator of what kind of record this is.")
@@ -316,7 +334,7 @@ class IEncodable(Interface):
         @type strio: File-like object
         @param strio: The stream to which to write bytes
 
-        @type compDict: C{dict} or C{None}
+        @type compDict: C{dict} or L{None}
         @param compDict: A dictionary of backreference addresses that have
         already been written to this stream and that may be used for
         compression.
@@ -499,13 +517,13 @@ class Name:
 
     def __eq__(self, other):
         if isinstance(other, Name):
-            return self.name == other.name
+            return self.name.lower() == other.name.lower()
         return NotImplemented
 
 
     def __ne__(self, other):
         if isinstance(other, Name):
-            return self.name != other.name
+            return self.name.lower() != other.name.lower()
         return NotImplemented
 
 
@@ -601,13 +619,13 @@ class _OPTHeader(tputil.FancyStrMixin, tputil.FancyEqMixin, object):
         attribute is a readonly property.
 
     @ivar type: The DNS record type. This is a fixed value of 41
-        (C{dns.OPT} for OPT Record. This attribute is a readonly
+        C{dns.OPT} for OPT Record. This attribute is a readonly
         property.
 
     @see: L{_OPTHeader.__init__} for documentation of other public
         instance attributes.
 
-    @see: L{https://tools.ietf.org/html/rfc6891#section-6.1.2}
+    @see: U{https://tools.ietf.org/html/rfc6891#section-6.1.2}
 
     @since: 13.2
     """
@@ -775,7 +793,7 @@ class _OPTVariableOption(tputil.FancyStrMixin, tputil.FancyEqMixin, object):
     @see: L{_OPTVariableOption.__init__} for documentation of public
         instance attributes.
 
-    @see: L{https://tools.ietf.org/html/rfc6891#section-6.1.2}
+    @see: U{https://tools.ietf.org/html/rfc6891#section-6.1.2}
 
     @since: 13.2
     """
@@ -880,22 +898,26 @@ class RRHeader(tputil.FancyEqMixin):
         @param cls: The query class.
 
         @type ttl: L{int}
-        @param ttl: Time to live for this record.
+        @param ttl: Time to live for this record.  This will be
+            converted to an L{int}.
 
         @type payload: An object implementing C{IEncodable}
         @param payload: A Query Type specific data object.
 
+        @raises TypeError: if the ttl cannot be converted to an L{int}.
         @raises ValueError: if the ttl is negative.
         """
         assert (payload is None) or isinstance(payload, UnknownRecord) or (payload.TYPE == type)
 
-        if ttl < 0:
+        integralTTL = int(ttl)
+
+        if integralTTL < 0:
             raise ValueError("TTL cannot be negative")
 
         self.name = Name(name)
         self.type = type
         self.cls = cls
-        self.ttl = ttl
+        self.ttl = integralTTL
         self.payload = payload
         self.auth = auth
 
@@ -1897,6 +1919,74 @@ class Record_MX(tputil.FancyStrMixin, tputil.FancyEqMixin):
 
 
 
+@implementer(IEncodable, IRecord)
+class Record_SSHFP(tputil.FancyEqMixin, tputil.FancyStrMixin):
+    """
+    A record containing the fingerprint of an SSH key.
+
+    @type algorithm: L{int}
+    @ivar algorithm: The SSH key's algorithm, such as L{ALGORITHM_RSA}.
+        Note that the numbering used for SSH key algorithms is specific
+        to the SSHFP record, and is not the same as the numbering
+        used for KEY or SIG records.
+
+    @type fingerprintType: L{int}
+    @ivar fingerprintType: The fingerprint type,
+        such as L{FINGERPRINT_TYPE_SHA256}.
+
+    @type fingerprint: L{bytes}
+    @ivar fingerprint: The key's fingerprint, e.g. a 32-byte SHA-256 digest.
+
+    @cvar ALGORITHM_RSA: The algorithm value for C{ssh-rsa} keys.
+    @cvar ALGORITHM_DSS: The algorithm value for C{ssh-dss} keys.
+    @cvar ALGORITHM_ECDSA: The algorithm value for C{ecdsa-sha2-*} keys.
+    @cvar ALGORITHM_Ed25519: The algorithm value for C{ed25519} keys.
+
+    @cvar FINGERPRINT_TYPE_SHA1: The type for SHA-1 fingerprints.
+    @cvar FINGERPRINT_TYPE_SHA256: The type for SHA-256 fingerprints.
+
+    @see: U{RFC 4255 <https://tools.ietf.org/html/rfc4255>}
+          and
+          U{RFC 6594 <https://tools.ietf.org/html/rfc6594>}
+    """
+    fancybasename = "SSHFP"
+    compareAttributes = ('algorithm', 'fingerprintType', 'fingerprint', 'ttl')
+    showAttributes = ('algorithm', 'fingerprintType', 'fingerprint')
+
+    TYPE = SSHFP
+
+    ALGORITHM_RSA = 1
+    ALGORITHM_DSS = 2
+    ALGORITHM_ECDSA = 3
+    ALGORITHM_Ed25519 = 4
+
+    FINGERPRINT_TYPE_SHA1 = 1
+    FINGERPRINT_TYPE_SHA256 = 2
+
+    def __init__(self, algorithm=0, fingerprintType=0, fingerprint=b'', ttl=0):
+        self.algorithm = algorithm
+        self.fingerprintType = fingerprintType
+        self.fingerprint = fingerprint
+        self.ttl = ttl
+
+
+    def encode(self, strio, compDict=None):
+        strio.write(struct.pack('!BB',
+                                self.algorithm, self.fingerprintType))
+        strio.write(self.fingerprint)
+
+
+    def decode(self, strio, length=None):
+        r = struct.unpack('!BB', readPrecisely(strio, 2))
+        (self.algorithm, self.fingerprintType) = r
+        self.fingerprint = readPrecisely(strio, length - 2)
+
+
+    def __hash__(self):
+        return hash((self.algorithm, self.fingerprintType, self.fingerprint))
+
+
+
 @implementer(IEncodable, IRecord)
 class Record_TXT(tputil.FancyEqMixin, tputil.FancyStrMixin):
     """
@@ -2000,13 +2090,98 @@ class Record_SPF(Record_TXT):
     @ivar data: Freeform text which makes up this record.
 
     @type ttl: L{int}
-    @ivar ttl: The maximum number of seconds which this record should be cached.
+    @ivar ttl: The maximum number of seconds
+               which this record should be cached.
     """
     TYPE = SPF
     fancybasename = 'SPF'
 
 
 
+@implementer(IEncodable, IRecord)
+class Record_TSIG(tputil.FancyEqMixin, tputil.FancyStrMixin):
+    """
+    A transaction signature, encapsulated in a RR, as described
+    in U{RFC 2845 <https://tools.ietf.org/html/rfc2845>}.
+
+    @type algorithm: L{Name}
+    @ivar algorithm: The name of the signature or MAC algorithm.
+
+    @type timeSigned: L{int}
+    @ivar timeSigned: Signing time, as seconds from the POSIX epoch.
+
+    @type fudge: L{int}
+    @ivar fudge: Allowable time skew, in seconds.
+
+    @type MAC: L{bytes}
+    @ivar MAC: The message digest or signature.
+
+    @type originalID: L{int}
+    @ivar originalID: A message ID.
+
+    @type error: L{int}
+    @ivar error: An error code (extended C{RCODE}) carried
+          in exceptional cases.
+
+    @type otherData: L{bytes}
+    @ivar otherData: Other data carried in exceptional cases.
+
+    """
+    fancybasename = "TSIG"
+    compareAttributes = ('algorithm', 'timeSigned', 'fudge',
+                         'MAC', 'originalID', 'error', 'otherData',
+                         'ttl')
+    showAttributes = ['algorithm', 'timeSigned', 'MAC', 'error', 'otherData']
+
+    TYPE = TSIG
+
+    def __init__(self, algorithm=None, timeSigned=None,
+                 fudge=5, MAC=None, originalID=0,
+                 error=OK, otherData=b'', ttl=0):
+        # All of our init arguments have to have defaults, because of
+        # the way IEncodable and Message.parseRecords() work, but for
+        # some of our arguments there is no reasonable default; we use
+        # invalid values here to prevent a user of this class from
+        # relying on what's really an internal implementation detail.
+        self.algorithm = None if algorithm is None else Name(algorithm)
+        self.timeSigned = timeSigned
+        self.fudge = str2time(fudge)
+        self.MAC = MAC
+        self.originalID = originalID
+        self.error = error
+        self.otherData = otherData
+        self.ttl = ttl
+
+
+    def encode(self, strio, compDict=None):
+        self.algorithm.encode(strio, compDict)
+        strio.write(struct.pack('!Q', self.timeSigned)[2:])  # 48-bit number
+        strio.write(struct.pack('!HH', self.fudge, len(self.MAC)))
+        strio.write(self.MAC)
+        strio.write(struct.pack('!HHH',
+                                self.originalID, self.error,
+                                len(self.otherData)))
+        strio.write(self.otherData)
+
+
+    def decode(self, strio, length=None):
+        algorithm = Name()
+        algorithm.decode(strio)
+        self.algorithm = algorithm
+        fields = struct.unpack('!QHH', b'\x00\x00' + readPrecisely(strio, 10))
+        self.timeSigned, self.fudge, macLength = fields
+        self.MAC = readPrecisely(strio, macLength)
+        fields = struct.unpack('!HHH', readPrecisely(strio, 6))
+        self.originalID, self.error, otherLength = fields
+        self.otherData = readPrecisely(strio, otherLength)
+
+
+    def __hash__(self):
+        return hash((self.algorithm, self.timeSigned,
+                     self.MAC, self.originalID))
+
+
+
 def _responseFromMessage(responseConstructor, message, **kwargs):
     """
     Generate a L{Message} like instance suitable for use as the response to
@@ -2035,6 +2210,43 @@ def _responseFromMessage(responseConstructor, message, **kwargs):
 
 
 
+def _getDisplayableArguments(obj, alwaysShow, fieldNames):
+    """
+    Inspect the function signature of C{obj}'s constructor,
+    and get a list of which arguments should be displayed.
+    This is a helper function for C{_compactRepr}.
+
+    @param obj: The instance whose repr is being generated.
+    @param alwaysShow: A L{list} of field names which should always be shown.
+    @param fieldNames: A L{list} of field attribute names which should be shown
+        if they have non-default values.
+    @return: A L{list} of displayable arguments.
+    """
+    displayableArgs = []
+    if _PY3:
+        # Get the argument names and values from the constructor.
+        signature = inspect.signature(obj.__class__.__init__)
+        for name in fieldNames:
+            defaultValue = signature.parameters[name].default
+            fieldValue = getattr(obj, name, defaultValue)
+            if (name in alwaysShow) or (fieldValue != defaultValue):
+                displayableArgs.append(' %s=%r' % (name, fieldValue))
+    else:
+        # Get the argument names and values from the constructor.
+        argspec = inspect.getargspec(obj.__class__.__init__)
+        # Reverse the args and defaults to avoid mapping positional arguments
+        # which don't have a default.
+        defaults = dict(zip(reversed(argspec.args), reversed(argspec.defaults)))
+        for name in fieldNames:
+            defaultValue = defaults.get(name)
+            fieldValue = getattr(obj, name, defaultValue)
+            if (name in alwaysShow) or (fieldValue != defaultValue):
+                displayableArgs.append(' %s=%r' % (name, fieldValue))
+
+    return displayableArgs
+
+
+
 def _compactRepr(obj, alwaysShow=None, flagNames=None, fieldNames=None,
                  sectionNames=None):
     """
@@ -2070,18 +2282,8 @@ def _compactRepr(obj, alwaysShow=None, flagNames=None, fieldNames=None,
         if name in alwaysShow or getattr(obj, name, False) == True:
             setFlags.append(name)
 
-    out = ['<', obj.__class__.__name__]
-
-    # Get the argument names and values from the constructor.
-    argspec = inspect.getargspec(obj.__class__.__init__)
-    # Reverse the args and defaults to avoid mapping positional arguments
-    # which don't have a default.
-    defaults = dict(zip(reversed(argspec.args), reversed(argspec.defaults)))
-    for name in fieldNames:
-        defaultValue = defaults.get(name)
-        fieldValue = getattr(obj, name, defaultValue)
-        if (name in alwaysShow) or (fieldValue != defaultValue):
-            out.append(' %s=%r' % (name, fieldValue))
+    displayableArgs = _getDisplayableArguments(obj, alwaysShow, fieldNames)
+    out = ['<', obj.__class__.__name__] + displayableArgs
 
     if setFlags:
         out.append(' flags=%s' % (','.join(setFlags),))
@@ -2367,7 +2569,7 @@ class Message(tputil.FancyEqMixin):
         """
         Retrieve the L{IRecord} implementation for the given record type.
 
-        @param type: A record type, such as L{A} or L{NS}.
+        @param type: A record type, such as C{A} or L{NS}.
         @type type: L{int}
 
         @return: An object which implements L{IRecord} or L{None} if none
@@ -2457,10 +2659,10 @@ class _EDNSMessage(tputil.FancyEqMixin, object):
         """
         Construct a new L{_EDNSMessage}
 
-        @see U{RFC1035 section-4.1.1<https://tools.ietf.org/html/rfc1035#section-4.1.1>}
-        @see U{RFC2535 section-6.1<https://tools.ietf.org/html/rfc2535#section-6.1>}
-        @see U{RFC3225 section-3<https://tools.ietf.org/html/rfc3225#section-3>}
-        @see U{RFC6891 section-6.1.3<https://tools.ietf.org/html/rfc6891#section-6.1.3>}
+        @see: U{RFC1035 section-4.1.1<https://tools.ietf.org/html/rfc1035#section-4.1.1>}
+        @see: U{RFC2535 section-6.1<https://tools.ietf.org/html/rfc2535#section-6.1>}
+        @see: U{RFC3225 section-3<https://tools.ietf.org/html/rfc3225#section-3>}
+        @see: U{RFC6891 section-6.1.3<https://tools.ietf.org/html/rfc6891#section-6.1.3>}
 
         @param id: A 16 bit identifier assigned by the program that generates
             any kind of query.  This identifier is copied the corresponding
@@ -2637,14 +2839,14 @@ class _EDNSMessage(tputil.FancyEqMixin, object):
     @classmethod
     def _fromMessage(cls, message):
         """
-        Construct and return a new L(_EDNSMessage} whose attributes and records
+        Construct and return a new L{_EDNSMessage} whose attributes and records
         are derived from the attributes and records of C{message} (a L{Message}
-        instance)
+        instance).
 
-        If present, an I{OPT} record will be extracted from the C{additional}
+        If present, an C{OPT} record will be extracted from the C{additional}
         section and its attributes and options will be used to set the EDNS
-        specific attributes C{extendedRCODE}, c{ednsVersion}, c{dnssecOK},
-        c{ednsOptions}.
+        specific attributes C{extendedRCODE}, C{ednsVersion}, C{dnssecOK},
+        C{ednsOptions}.
 
         The C{extendedRCODE} will be combined with C{message.rCode} and assigned
         to C{self.rCode}.
